/*
libfawk - A function-only AWK dialect

Copyright (c) 2017..2019 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/

#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include <stdarg.h>
#include <genht/hash.h>
#include "array.h"
#include "str.h"
#include "execute.h"
#include "symtab.h"

static unsigned int arrhash(const void *k)
{
	const fawk_arridx_t *i = k;
	switch(i->type) {
		case FAWK_NUM: return i->data.num;
		case FAWK_STRNUM: return i->data.str->num;
		case FAWK_STR: return strhash(i->data.str->str);
		case FAWK_NIL: return strhash("\001NIL\001");
		default: abort();
	}
}

#define keyconv(i, n, s, isnum) \
	if (i->type == FAWK_NUM)          { isnum = 1; n = i->data.num; } \
	else if (i->type == FAWK_STRNUM)  { isnum = 1; n = i->data.str->num; } \
	else if (i->type == FAWK_NIL)     { isnum = 0; s = "\001NIL\001"; } \
	else { s = i->data.str->str; }

static int arrkeyeq(const void *k1, const void *k2)
{
	const fawk_arridx_t *i1 = k1, *i2 = k2;
	fawk_num_t n1, n2;
	int isnum1 = 0, isnum2 = 0;
	const char *s1 = NULL, *s2 = NULL;

	keyconv(i1, n1, s1, isnum1);
	keyconv(i2, n2, s2, isnum2);

	if (isnum1 && isnum2)
		return n1 == n2;

	if (!isnum1 && !isnum2)
		return !strcmp(s1, s2);

	return 0;
}


FAWK_API void fawk_array_init(fawk_ctx_t *ctx, fawk_cell_t *dst)
{
	dst->data.arr = fawk_calloc(ctx, sizeof(fawk_arr_t), 1); if (dst->data.arr == NULL) { dst->type = FAWK_NIL; return; }
	dst->type = FAWK_ARRAY;
	dst->data.arr->uid = ctx->arr_uid++;
	dst->data.arr->refco = 1;
	dst->data.arr->destroying = 0;
	htpp_init(&dst->data.arr->hash, arrhash, arrkeyeq);
}

FAWK_API void fawk_array_free(fawk_ctx_t *ctx, fawk_cell_t *dst)
{
	if (dst->data.arr->destroying) /* self-referencing array: a[1] = a */
		return;
	dst->data.arr->destroying = 1;
	dst->data.arr->refco--;
	if (dst->data.arr->refco == 0) {
		htpp_entry_t *e;
		for (e = htpp_first(&dst->data.arr->hash); e; e = htpp_next(&dst->data.arr->hash, e)) {
			fawk_arridx_t *idx = e->key;
			if ((idx->type == FAWK_STR) || (idx->type == FAWK_STRNUM))
				fawk_str_free(ctx, idx->data.str);
			fawk_cell_free(ctx, e->value);
			fawk_free(ctx, e->value);
			fawk_free(ctx, e->key);
		}
		htpp_uninit(&dst->data.arr->hash);
		fawk_free(ctx, dst->data.arr); dst->data.arr = NULL;
		dst->type = FAWK_NIL;
	}
	else
		dst->data.arr->destroying = 0;
}


FAWK_API fawk_arridx_t *fawk_array_dump_list(fawk_ctx_t *ctx, fawk_cell_t *arrcell, size_t *out_len)
{
	size_t len, n;
	fawk_arridx_t *list;
	htpp_entry_t *e;

	if (arrcell->type != FAWK_ARRAY) return NULL;
	len = arrcell->data.arr->hash.used;
	list = fawk_malloc(ctx, sizeof(fawk_arridx_t) * len); if (list == NULL) return NULL;

	for (e = htpp_first(&arrcell->data.arr->hash), n = 0; e; e = htpp_next(&arrcell->data.arr->hash, e), n++) {
		fawk_arridx_t *idx = e->key;
		list[n].type = idx->type;
		if ((idx->type == FAWK_STR) || (idx->type == FAWK_STRNUM)) {
			list[n].data.str = fawk_str_dup(ctx, idx->data.str);
			if (list[n].data.str == NULL) { list[n].type = FAWK_NIL; ctx->exec.error = 1; }
		}
		else if (idx->type != FAWK_NIL) list[n].data.num = idx->data.num;
	}
	*out_len = len;
	return list;
}

FAWK_API fawk_cell_t *fawk_array_resolve_c(fawk_ctx_t *ctx, int create, fawk_cell_t *arrcell, ...)
{
	fawk_symref_t sr;
	fawk_arridx_t idx[64];
	va_list ap;
	int n;

	sr.ref.global = arrcell; sr.is_local = 0;
	sr.idx_len = 1; sr.idx = idx;

	va_start(ap, arrcell);
	for(n = 0; n < sizeof(idx)/sizeof(idx[0]); n++) {
		idx[n].type = va_arg(ap, fawk_celltype_t);
		switch(idx[n].type) {
			case FAWK_NIL: va_end(ap); return fawk_symtab_deref(ctx, &sr, create, NULL);
			case FAWK_NUM: idx[n].data.num = va_arg(ap, fawk_num_t); break;
			case FAWK_STR: idx[n].data.str = fawk_str_new_from_literal(ctx, va_arg(ap, char *), -1); if (idx[n].data.str == NULL) goto err; break;
			default: goto err;
		}
	}

	err:; va_end(ap);
	/* too many indices */
	return NULL;
}
