/*
libfawk - A function-only AWK dialect

Copyright (c) 2017..2019 Tibor 'Igor2' Palinkas. All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  * 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
  * 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
  * 3. Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from this
    software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

Project page: http://repo.hu/projects/libfawk
Source code: svn://repo.hu/libfawk/trunk
Contact the author: http://igor2.repo.hu/contact.html
*/

#include <stdio.h>
#include "libfawk.h"
#include "builtin.h"
#include "str.h"
#include "symtab.h"
#include "execute.h"
#include "cast.h"

static void fawk_bi_int(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
	if (argc == 1) {
		fawk_cell_cpy(ctx, retval, FAWK_CFUNC_ARG(0));
		fawk_cast_to_num(ctx, retval);
		retval->data.num = (int)retval->data.num;
	}
}

static void fawk_bi_length(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
	if (argc == 1) {
		fawk_cell_t *v = FAWK_CFUNC_ARG(0);

		switch(v->type) {
			case FAWK_STR:
			case FAWK_STRNUM:
				retval->data.num = v->data.str->used;
				break;
			case FAWK_ARRAY:
				retval->data.num = v->data.arr->hash.used;
				break;
			case FAWK_NIL:
			case FAWK_NUM: /* there is no reason to do anything for num: it can be casted to str in the call */
			case FAWK_FUNC:
			case FAWK_SYMREF:
			case FAWK_CCALL_RET:
				return; /* NIL for invalid types */
		}
		retval->type = FAWK_NUM;
	}
}

static void fawk_bi_delete(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
	int n;
	for(n = 0; n < argc; n++) {
		fawk_cell_t *arr, *item, *v = FAWK_CFUNC_ARG(n);
		if (v->type == FAWK_SYMREF) {
			item = fawk_symtab_deref(ctx, &v->data.symref, 0, &arr);
			if (item == NULL)
				continue;
			if (arr != NULL)
				htpp_pop(&arr->data.arr->hash, &v->data.symref.idx[v->data.symref.idx_len-1]);
			fawk_cell_free(ctx, item);
		}
	}
}

static void fawk_bi_isarray(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
	fawk_cell_t *arr, *v = FAWK_CFUNC_ARG(0);

	retval->type = FAWK_NUM; retval->data.num = 0;
	if (v->type == FAWK_SYMREF) {
		if (fawk_symtab_deref(ctx, &v->data.symref, 0, &arr) == NULL) return;
		if (arr != NULL) retval->data.num = 1;
	} else if (v->type == FAWK_ARRAY) retval->data.num = 1;
}

static void fawk_bi_print_cell(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
#ifndef FAWK_DISABLE_FAWK_PRINT
	int n;
	for(n = 0; n < argc; n++) {
		fawk_dump_cell(FAWK_CFUNC_ARG(n), fname[10] == '_');
		printf(n == argc-1 ? "\n" : " ");
	}
#endif
}


static void fawk_bi_substr(fawk_ctx_t *ctx, const char *fname, int argc, fawk_cell_t *retval)
{
	fawk_cell_t *str = FAWK_CFUNC_ARG(0), *from = FAWK_CFUNC_ARG(1), *len, dummy;

	if ((argc != 2) && (argc != 3)) return;

	fawk_cast_to_str(ctx, str); fawk_cast_to_num(ctx, from);
	if (argc > 2) { len = FAWK_CFUNC_ARG(2); fawk_cast_to_num(ctx, len); } else { len = &dummy; len->data.num = str->data.str->used; }

	if (--from->data.num < 0) from->data.num = 0;
	if (from->data.num > str->data.str->used) from->data.num = str->data.str->used;

	retval->type = FAWK_STR; retval->data.str = fawk_str_new_from_literal(ctx, str->data.str->str + (long)from->data.num, len->data.num);
}

FAWK_API int fawk_builtin_init(fawk_ctx_t *ctx)
{
	fawk_cell_t *vs = fawk_symtab_regvar(ctx, "SUBSEP", FAWK_SCALAR), *vv = fawk_symtab_regvar(ctx, "FAWK_API_VER", FAWK_SCALAR);
	if ((vs == NULL) || (vv == NULL)) return -1;
	vs->type = FAWK_STR; vs->data.str = fawk_str_new_from_literal(ctx, "\034", -1); if (vs->data.str == NULL) return -1;
	vv->type = FAWK_NUM; vv->data.num = FAWK_API_VER;

	fawk_symtab_regcfunc(ctx, "int", fawk_bi_int);
	fawk_symtab_regcfunc(ctx, "length", fawk_bi_length);
	fawk_symtab_regcfunc(ctx, "delete", fawk_bi_delete);
	fawk_symtab_regcfunc(ctx, "isarray", fawk_bi_isarray);
	fawk_symtab_regcfunc(ctx, "fawk_print_cell", fawk_bi_print_cell);
	fawk_symtab_regcfunc(ctx, "fawk_print", fawk_bi_print_cell);
	fawk_symtab_regcfunc(ctx, "substr", fawk_bi_substr);
	return 0;
}
